from dataclasses import dataclass, field
from functools import cached_property, partial
from typing import Iterable, List, Union

from beet import Context, Function, FunctionTag
from bolt import Runtime
from mecha import Mecha
from pydantic import BaseModel

from . import resolver
from .ast import ConstantScoreChecker, ObjectiveChecker, SourceJsonConverter
from .literals import literal_types
from .node import ExpressionNode
from .operations import (
    Append,
    GenericValue,
    Insert,
    Merge,
    MergeRoot,
    Operation,
    Prepend,
    Set,
    wrapped_max,
    wrapped_min,
)
from .optimizer import Optimizer
from .sources import (
    ConstantScoreSource,
    DataSource,
    ScoreSource,
    Source,
    TempScoreSource,
)
from .utils import identifier_generator

# from rich import print
# from rich.pretty import pprint

__all__ = [
    "ExpressionOptions",
    "Expression",
    "Scoreboard",
    "Score",
    "Data",
]


class ExpressionOptions(BaseModel):
    """Bolt Expressions Options"""

    temp_objective: str = "bolt.expr.temp"
    const_objective: str = "bolt.expr.const"
    temp_storage: str = "bolt.expr:temp"
    init_path: str = "init_expressions"
    objective_prefix: str = ""

    disable_commands: bool = False


@dataclass
class Expression:
    ctx: Context
    activated: bool = False
    called_init: bool = False
    init_commands: List[str] = field(default_factory=list)

    def __post_init__(self):
        if not self.activated:
            self.opts = self.ctx.validate("bolt_expressions", ExpressionOptions)
            self._runtime.expose(
                "min", partial(wrapped_min, self._runtime.globals.get("min", min))
            )
            self._runtime.expose(
                "max", partial(wrapped_max, self._runtime.globals.get("max", max))
            )
            self.activated = True

        if not self.opts.disable_commands:
            self.ctx.require("bolt_expressions.contrib.commands")

        helpers = self._runtime.helpers

        helpers["interpolate_json"] = SourceJsonConverter(helpers["interpolate_json"])

        Set.on_resolve(self.resolve)
        TempScoreSource.objective = self.opts.temp_objective
        ConstantScoreSource.objective = self.opts.const_objective

    @cached_property
    def _runtime(self) -> Runtime:
        return self.ctx.inject(Runtime)

    @cached_property
    def _mc(self) -> Mecha:
        return self.ctx.inject(Mecha)

    def _inject_command(self, cmd: str):
        self._runtime.commands.append(self._mc.parse(cmd, using="command"))

    def resolve(self, nodes: Operation):
        # pprint(nodes)
        nodes = list(nodes.unroll())
        # pprint(nodes)
        nodes = list(Optimizer.optimize(nodes))
        # pprint(nodes)
        cmds = list(resolver.resolve(nodes))
        # pprint(cmds, expand_all=True)
        for cmd in cmds:
            self._inject_command(cmd)

    def set(self, source: Source, value: GenericValue):
        Set.create(source, value).resolve()

    def init(self):
        """Injects a function which creates `ConstantSource` fakeplayers"""
        path = self.ctx.generate.path(self.opts.init_path)
        self._inject_command(f"function {path}")
        self.called_init = True

    def generate_init(self):
        if not self.init_commands:
            return

        self.ctx.generate(
            self.opts.init_path,
            Function(
                self.init_commands,
                prepend_tags=["minecraft:load"] if not self.called_init else None,
            ),
        )


@dataclass
class Scoreboard:
    """API for manipulating scoreboards.

    To use, inject the current `Context` and construct an `Score` instance.
    ```
        Objective = ctx.inject(Scoreboard)
        my_obj = Objective["my_obj"]
    ```
    Now you can perform the API manipulation via the operators:
    ```
        my_obj["@s"] += 10
        my_obj["temp"] = my_obj["@s"] * 10
        player = my_obj["@s"]
        player += 10 * my_obj["temp"]
    ```
    """

    ctx: Context
    objectives: set[str] = field(default_factory=set)
    constants: set[int] = field(default_factory=set)

    added_objectives: set[str] = field(init=False, default_factory=set)

    def __post_init__(self):
        self._expr = self.ctx.inject(Expression)

        opts = self._expr.opts

        self.objectives.update((opts.const_objective, opts.temp_objective))

        self._expr._mc.check.extend(
            ConstantScoreChecker(
                objective=opts.const_objective, callback=self.add_constant
            ),
            ObjectiveChecker(
                whitelist=self.objectives,
                callback=self.add_objective,
            ),
        )
        ScoreSource.on_rebind(self.set_score)
        ScoreSource.attach("reset", self.reset)
        ScoreSource.attach("enable", self.enable)

    def add_objective(self, name: str, criteria: str = "dummy"):
        if name not in self.added_objectives:
            self.added_objectives.add(name)

            self._expr.init_commands.insert(
                0, f"scoreboard objectives add {name} {criteria}"
            )

    def add_constant(self, node: ConstantScoreSource):
        if not node.value in self.constants:
            self.constants.add(node.value)
            self._expr.init_commands.append(
                f"scoreboard players set {node} {node.value}"
            )

    def set_score(self, score: ScoreSource, value: GenericValue):
        return self._expr.set(score, value)

    def objective(self, name: str, criteria: str = None, prefixed=True):
        """Get a Score instance through the Scoreboard API"""
        if prefixed:
            name = self._expr.opts.objective_prefix + name

        if criteria:
            self.add_objective(name, criteria)

        return Score(self, name)

    def __call__(self, objective: str, criteria: str = None, prefixed=True):
        return self.objective(objective, criteria, prefixed)

    def reset(self, source: ScoreSource):
        cmd = resolver.generate("reset:score", source)
        self._expr._inject_command(cmd)

    def enable(self, source: ScoreSource):
        cmd = resolver.generate("enable:score", source)
        self._expr._inject_command(cmd)


@dataclass
class Score:
    ref: Scoreboard = field(repr=False)
    objective: str

    def __getitem__(self, scoreholder: Union[str, List[str]]) -> ScoreSource:
        if type(scoreholder) is str:
            return ScoreSource.create(scoreholder, self.objective)
        return [ScoreSource.create(holder, self.objective) for holder in scoreholder]

    def __setitem__(self, scoreholder: str, value: Operation):
        self.ref.set_score(self[scoreholder], value)

    def __str__(self):
        return self.objective


@dataclass
class Data:
    ctx: Context

    def __post_init__(self):
        self._expr = self.ctx.inject(Expression)
        self.identifiers = identifier_generator(self.ctx)
        DataSource.on_rebind(self.set_data)
        DataSource.attach("remove", self.remove)
        DataSource.attach("append", self.append)
        DataSource.attach("prepend", self.prepend)
        DataSource.attach("insert", self.insert)
        DataSource.attach("merge", self.merge)

    def __call__(self, target: str):
        """Guess target type and return a data source."""
        ...

    def set_data(self, source: DataSource, value: GenericValue):
        return self._expr.set(source, value)

    def storage(self, resource_location: str):
        return DataSource.create("storage", resource_location)

    def entity(self, entity: str):
        return DataSource.create("entity", entity)

    def block(self, position: str):
        return DataSource.create("block", position)

    def dummy(self, scale: int = 1, type: str = "int"):
        "Create a dummy data source in a storage."
        path = next(self.identifiers)
        target = self._expr.opts.temp_storage
        return DataSource.create("storage", target, path, scale, type)

    def cast(self, value: Union[Source, Operation], type: str):
        source = self.dummy(type=type)
        if not isinstance(value, ExpressionNode):
            value = literal_types[type](value)
        self._expr.set(source, value)
        return source._copy(nbt_type=None)

    def remove(self, source: DataSource, value: Union[str, int] = None):
        node = source if value is None else source[value]
        if not len(node._path):
            raise ValueError(
                f'Cannot remove the root of {node._type} "{node._target}".'
            )
        cmd = resolver.generate("remove:data", value=node)
        self._expr._inject_command(cmd)

    def append(self, source: DataSource, value: GenericValue):
        self._expr.resolve(Append.create(source, value))

    def prepend(self, source: DataSource, value: GenericValue):
        self._expr.resolve(Prepend.create(source, value))

    def insert(self, source: DataSource, index: int, value: GenericValue):
        self._expr.resolve(Insert.create(source, value, index=index))

    def merge(self, source: DataSource, value: GenericValue):
        Operation = Merge if len(source._path) else MergeRoot
        self._expr.resolve(Operation.create(source, value))



from dataclasses import dataclass, field
from typing import Callable, Iterable, Optional, Set, overload

__all__ = [
    "ExpressionNode",
]


@dataclass(unsafe_hash=True, order=False, eq=False)
class ExpressionNode:
    attached_methods = set()

    @classmethod
    def link(cls, magic_method: str, reverse=False):
        def decorator(operation_class):
            def normal(self, other: "ExpressionNode"):
                return operation_class.create(self, other)

            def reversed(self, other: "ExpressionNode"):
                return operation_class.create(other, self)

            setattr(cls, f"__{magic_method}__", normal)
            if reverse:
                setattr(cls, f"__r{magic_method}__", reversed)
            return operation_class

        return decorator

    @classmethod
    def attach(cls, method_name: str, function: Callable):
        """Attach a method to an expression node."""
        if getattr(cls, method_name, None) and method_name not in cls.attached_methods:
            return

        def method(self, *args, **kwargs):
            return function(self, *args, **kwargs)

        setattr(cls, method_name, method)
        cls.attached_methods.add(method_name)

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def unroll(self):
        yield self



from typing import Dict, Iterable, List

from .operations import GenericValue, Operation
from .sources import DataSource, ScoreSource, Source

__all__ = [
    "get_templates",
    "get_type",
    "generate",
    "resolve_execute",
    "resolve_execute_store",
    "resolve_set_data_data",
    "resolve_node",
    "resolve",
]

Command: type = str


def get_templates() -> Dict[str, str]:
    return {
        "node": resolve_node,
        "execute": resolve_execute,
        "execute:store": resolve_execute_store,
        "set:score:literal": lambda op: f"scoreboard players set {op.former} {op.latter}",
        "add:score:literal": lambda op: f"scoreboard players add {op.former} {op.latter}",
        "subtract:score:literal": lambda op: f"scoreboard players remove {op.former} {op.latter}",
        "set:score:score": lambda op: f"scoreboard players operation {op.former} = {op.latter}",
        "add:score:score": lambda op: f"scoreboard players operation {op.former} += {op.latter}",
        "subtract:score:score": lambda op: f"scoreboard players operation {op.former} -= {op.latter}",
        "multiply:score:score": lambda op: f"scoreboard players operation {op.former} *= {op.latter}",
        "divide:score:score": lambda op: f"scoreboard players operation {op.former} /= {op.latter}",
        "modulus:score:score": lambda op: f"scoreboard players operation {op.former} %= {op.latter}",
        "min:score:score": lambda op: f"scoreboard players operation {op.former} < {op.latter}",
        "max:score:score": lambda op: f"scoreboard players operation {op.former} > {op.latter}",
        "set:data:literal": lambda op: f"data modify {op.former} set value {op.latter}",
        "set:data:data": resolve_set_data_data,
        "set:data:score": lambda op: f"execute store result {op.former} {op.former.get_type()} {op.former._scale} run scoreboard players get {op.latter}",
        "set:score:data": lambda op: f"execute store result score {op.former} run data get {op.latter} {op.latter._scale}",
        "append:data:literal": lambda op: f"data modify {op.former} append value {op.latter}",
        "append:data:data": lambda op: f"data modify {op.former} append from {op.latter}",
        "prepend:data:literal": lambda op: f"data modify {op.former} prepend value {op.latter}",
        "prepend:data:data": lambda op: f"data modify {op.former} prepend from {op.latter}",
        "insert:data:literal": lambda op: f"data modify {op.former} insert {op.index} value {op.latter}",
        "insert:data:data": lambda op: f"data modify {op.former} insert {op.index} from {op.latter}",
        "merge:data:literal": lambda op: f"data modify {op.former} merge value {op.latter}",
        "merge:data:data": lambda op: f"data modify {op.former} merge from {op.latter}",
        "mergeroot:data:literal": lambda op: f"data merge {op.former} {op.latter}",
        "remove:data": lambda source: f"data remove {source}",
        "reset:score": lambda source: f"scoreboard players reset {source}",
        "enable:score": lambda source: f"scoreboard players enable {source}",
    }


def get_type(node: GenericValue):
    # optimizer might convert an int score back to a literal int
    # for operations like Set, Add and Subtract
    if isinstance(node, ScoreSource):
        return "score"
    if isinstance(node, DataSource):
        return "data"
    return "literal"


def generate(template_id: str, *args, **kwargs):
    template = get_templates()[template_id]
    return template(*args, *kwargs.values())


def resolve_execute(node: Operation):
    args = []
    for source in node.store:
        args.append(generate("execute:store", source))
    return " ".join(("execute", *args, "run ")) if args else ""


def resolve_execute_store(source: Source):
    if isinstance(source, ScoreSource):
        return f"store result score {source}"
    if isinstance(source, DataSource):
        return f"store result {source} {source.get_type()} {source._scale}"


def resolve_set_data_data(op: Operation):
    if (
        op.former._scale != 1
        or op.latter._scale != 1
        or op.former._nbt_type is not None
    ):
        return f"execute store result {op.former} {op.former.get_type()} {op.former._scale} run data get {op.latter} {op.latter._scale}"
    return f"data modify {op.former} set from {op.latter}"


def resolve_node(node: Operation):
    id = node.__class__.__name__.lower()  # TODO Operation should have an id property
    former_type = get_type(node.former)
    latter_type = get_type(node.latter)
    template_id = f"{id}:{former_type}:{latter_type}"
    cmd = generate("execute", node)
    return cmd + generate(template_id, node)


def resolve(nodes: List[Operation]) -> Iterable[Command]:
    """Transforms a list of operation nodes into command strings."""
    yield from map(resolve_node, nodes)



__version__ = "0.12.1"

from .api import Data as _Data
from .api import Expression as _Expression
from .api import Scoreboard as _Scoreboard
from .api import *
from .ast import *
from .literals import *
from .node import *
from .operations import *
from .optimizer import *
from .plugin import *
from .resolver import *
from .sources import *
from .utils import *


def __monkey_patch():
    from nbtlib import Compound, List

    def __patched_hash(self):
        return hash(str(self))

    Compound.__hash__ = __patched_hash
    List.__hash__ = __patched_hash


__monkey_patch()



from dataclasses import dataclass, field
from typing import Any, Callable, Iterable, List, Tuple, Union

from nbtlib import Int

from .literals import Literal
from .node import ExpressionNode
from .sources import ConstantScoreSource, DataSource, Source, TempScoreSource

# from rich import print
# from rich.pretty import pprint

__all__ = [
    "Operation",
    "DataOperation",
    "Merge",
    "MergeRoot",
    "Insert",
    "Append",
    "Prepend",
    "Set",
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "Modulus",
    "Min",
    "Max",
    "wrapped_min",
    "wrapped_max",
]

GenericValue = Union["Operation", "Source", int, str]


def wrapped_min(f, *args, **kwargs):
    values = args

    if len(args) == 1:
        if not isinstance(args[0], Iterable):
            return args[0]

        values = tuple(args[0])

    for i, node in enumerate(values):
        if not isinstance(node, ExpressionNode):
            continue

        remaining = values[:i] + values[i + 1 :]
        return Min.create(wrapped_min(f, *remaining, **kwargs), node)

    return f(*args, **kwargs)


def wrapped_max(f, *args, **kwargs):
    values = args

    if len(args) == 1:
        if not isinstance(args[0], Iterable):
            return args[0]

        values = tuple(args[0])

    for i, node in enumerate(values):
        if not isinstance(node, ExpressionNode):
            continue

        remaining = values[:i] + values[i + 1 :]
        return Max.create(wrapped_max(f, *remaining, **kwargs), node)

    return f(*args, **kwargs)


@dataclass(unsafe_hash=False, order=False)
class Operation(ExpressionNode):
    former: GenericValue
    latter: GenericValue
    store: Tuple[Source] = field(default_factory=tuple)

    @classmethod
    def create(cls, former: GenericValue, latter: GenericValue, *args, **kwargs):
        """Factory method to create new operations"""

        if not isinstance(former, ExpressionNode):
            former = Literal.create(former)
        if not isinstance(latter, ExpressionNode):
            latter = Literal.create(latter)

        return super().create(former, latter, *args, **kwargs)

    def unroll(self) -> Iterable["Operation"]:
        *former_nodes, former_var = self.former.unroll()
        *latter_nodes, latter_var = self.latter.unroll()

        yield from former_nodes
        yield from latter_nodes

        if type(former_var) is TempScoreSource:
            temp_var = former_var
        else:
            temp_var = TempScoreSource.create()
            yield Set.create(temp_var, former_var)
        yield self.__class__.create(temp_var, latter_var)
        yield temp_var


class DataOperation(Operation):
    ...


class Merge(DataOperation):
    def unroll(self):
        yield self


class MergeRoot(Merge):
    ...


@dataclass(unsafe_hash=False, order=False)
class Insert(DataOperation):
    index: int = 0

    def unroll(self):
        if type(self.latter) in (DataSource, Literal):
            yield self
        else:
            *latter_nodes, latter_var = self.latter.unroll()
            yield from latter_nodes
            yield self.__class__.create(self.former, 0, index=self.index)
            yield Set.create(self.former[self.index], latter_var)


class Append(Insert):
    @classmethod
    def create(cls, former: DataSource, latter: GenericValue, *args, **kwargs):
        return super().create(former, latter, index=-1)


class Prepend(Insert):
    @classmethod
    def create(cls, former: DataSource, latter: GenericValue, *args, **kwargs):
        return super().create(former, latter, index=0)


class Set(Operation):
    @classmethod
    def on_resolve(cls, callback: Callable):
        cls._resolve = callback

    def unroll(self) -> Iterable["Operation"]:
        TempScoreSource.count = -1

        if type(self.latter) is DataSource:
            yield Set.create(self.former, self.latter)
        else:
            *latter_nodes, latter_var = self.latter.unroll()
            yield from latter_nodes
            yield Set.create(self.former, latter_var)

    def resolve(self):
        return self._resolve(self)


class ScoreOperation(Operation):
    ...


@ExpressionNode.link("add", reverse=True)
class Add(ScoreOperation):
    @classmethod
    def create(cls, former: GenericValue, latter: GenericValue):
        if not isinstance(former, Operation) and isinstance(latter, Operation):
            return super().create(latter, former)
        return super().create(former, latter)


@ExpressionNode.link("sub", reverse=True)
class Subtract(ScoreOperation):
    ...  # fmt: skip


@ExpressionNode.link("mul", reverse=True)
class Multiply(ScoreOperation):
    @classmethod
    def create(cls, former: GenericValue, latter: GenericValue):
        if not isinstance(former, Operation) and isinstance(latter, Operation):
            return super().create(latter, former)
        return super().create(former, latter)


@ExpressionNode.link("truediv", reverse=True)
class Divide(ScoreOperation):
    ...


@ExpressionNode.link("mod", reverse=True)
class Modulus(ScoreOperation):
    ...


@ExpressionNode.link("min", reverse=True)
class Min(ScoreOperation):
    ...


@ExpressionNode.link("max", reverse=True)
class Max(ScoreOperation):
    ...


@ExpressionNode.link("if")
class If(Operation):
    ...


# @ExpressionNode.link("eq")
# class Equal(Operation): ...


# @ExpressionNode.link("ne")
# class NotEqual(Operation): ...


@ExpressionNode.link("lt")
class LessThan(Operation):
    ...


@ExpressionNode.link("gt")
class GreaterThan(Operation):
    ...


@ExpressionNode.link("le")
class LessThanOrEqualTo(Operation):
    ...


@ExpressionNode.link("ge")
class GreaterThanOrEqualTo(Operation):
    ...


@ExpressionNode.link("abs")
class Abs(Operation):
    @classmethod
    def create(cls, former: GenericValue):
        return If.create(LessThan.create(former, 0), Multiply.create(former, -1))


# def __neg__(self):
#     return Multiply.create(self, -1)

# def __pos__(self):
#     return self

# def __abs__(self):
#     return If.create(LessThan.create(self, 0), Multiply.create(self, -1))



from dataclasses import dataclass, field
from functools import cache
from itertools import count
from typing import Callable, ClassVar, Union

from nbtlib import Compound, Path

from . import operations as op
from .literals import convert_tag
from .node import ExpressionNode

# from rich.pretty import pprint

__all__ = [
    "Source",
    "ScoreSource",
    "ConstantScoreSource",
    "TempScoreSource",
    "DataSource",
    "parse_compound",
]

SOLO_COLON = slice(None, None, None)


class Source(ExpressionNode):
    ...


@dataclass(unsafe_hash=True, order=False)
class ScoreSource(Source):
    scoreholder: str
    objective: str

    @classmethod
    def on_rebind(cls, callback: Callable):
        cls._rebind = callback

    def __rebind__(self, other: ExpressionNode):
        self._rebind(self, other)
        return self

    def __str__(self):
        return f"{self.scoreholder} {self.objective}"

    def __repr__(self):
        return f'"{str(self)}"'

    def component(self, **tags):
        return {
            "score": {"name": self.scoreholder, "objective": self.objective},
            **tags,
        }

    @property
    def holder(self):
        return self.scoreholder

    @property
    def obj(self):
        return self.objective


@dataclass(unsafe_hash=True, order=False)
class ConstantScoreSource(ScoreSource):
    objective: str = "const"
    value: int = field(hash=False, kw_only=True)

    @classmethod
    def create(cls, value: int):
        return super().create(f"${value}", cls.objective, value=value)


class TempScoreSource(ScoreSource):
    objective: str = "temp"
    count: ClassVar[int] = -1

    @classmethod
    def create(cls):
        cls.count += 1
        return super().create(f"$i{cls.count}", cls.objective)


def parse_compound(value: Union[str, dict, Path, Compound]):
    if type(value) in (Path, Compound):
        return value
    if type(value) is dict:
        return convert_tag(value)
    return Path(value)


@dataclass(unsafe_hash=True, order=False)
class DataSource(Source):
    _default_nbt_type: ClassVar[str] = "int"
    _default_floating_point_type: ClassVar[str] = "double"
    _type: str
    _target: str
    _path: Path = field(default_factory=Path)
    _scale: float = 1
    _nbt_type: str = None

    _constructed: bool = field(hash=False, default=False)

    def __post_init__(self):
        self._constructed = True

    @classmethod
    def on_rebind(cls, callback: Callable):
        cls._rebind = callback

    def unroll(self):
        temp_var = TempScoreSource.create()
        yield op.Set.create(temp_var, self)
        yield temp_var

    def __rebind__(self, other):
        self._rebind(self, other)
        return self

    def __setattr__(self, key: str, value):
        if not self._constructed:
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def __setitem__(self, key: str, value):
        child = self.__getitem__(key)
        child._rebind(child, value)

    def __getitem__(self, key: Union[str, int, Path, Compound]) -> "DataSource":
        if key is SOLO_COLON:
            # self[:]
            return self.all()
        if (
            isinstance(key, dict)
            or isinstance(key, str)
            and (key[0], key[-1]) == ("{", "}")
        ):
            # self[{abc:1b}]
            return self.filtered(key)
        # self[0] or self.foo
        path = self._path[key]
        return self._copy(path=path)

    __getattr__ = __getitem__

    def __call__(
        self,
        matching: Union[str, Path, Compound] = None,
        scale: float = None,
        type: str = None,
    ) -> "DataSource":
        """Create a new DataSource with modified properties."""
        if matching is not None:
            path = self._path[parse_compound(matching)]
        else:
            path = self._path
        return self._copy(
            path=path,
            scale=scale if scale is not None else self._scale,
            nbt_type=type if type is not None else self._nbt_type,
        )

    def __str__(self):
        return f"{self._type} {self._target} {self._path}"

    def __repr__(self):
        return f'"{str(self)}"'

    def _copy(self, **kwargs) -> "DataSource":
        """Create a new DataSource with overwritten properties."""
        return self.__class__.create(
            _type=kwargs.get("type", self._type),
            _target=kwargs.get("target", self._target),
            _path=kwargs.get("path", self._path),
            _scale=kwargs.get("scale", self._scale),
            _nbt_type=kwargs.get("nbt_type", self._nbt_type),
        )

    def get_type(self):
        return self._nbt_type if self._nbt_type else self._default_nbt_type

    def all(self) -> "DataSource":
        path = self._path + "[]"
        return self._copy(path=path)

    def filtered(self, value: Union[str, Path, Compound]):
        compound = parse_compound(value)
        path = self._path[:][compound]
        return self._copy(path=path)

    def component(self, **tags):
        return {"nbt": str(self._path), self._type: self._target, **tags}



from typing import Dict

from beet import Context
from bolt import Runtime

__all__ = [
    "identifier_generator",
]


def identifier_generator(ctx: Context):
    runtime = ctx.inject(Runtime)
    incr: Dict[str, int] = {}
    while True:
        path = runtime.modules.current_path
        incr[path] = incr.setdefault(path, -1) + 1
        yield ctx.generate.format(f"{{hash}}_{incr[path]}", path)



from dataclasses import dataclass
from typing import ClassVar

from nbtlib import Base, Byte, Compound, Double, Float, Int, List, Long, Short, String

from .node import ExpressionNode

__all__ = [
    "Literal",
    "literal_types",
    "convert_tag",
]

literal_types = {
    "byte": Byte,
    "short": Short,
    "int": Int,
    "long": Long,
    "float": Float,
    "double": Double,
    "compound": Compound,
    "list": List,
    "string": String,
}


def convert_tag(value):
    if isinstance(value, Base):
        return value
    t = type(value)
    if t is list:
        return List([convert_tag(x) for x in value])
    if t is dict:
        return Compound({key: convert_tag(value) for key, value in value.items()})
    if t is bool:
        return Byte(value)
    if t is int:
        return Int(value)
    if t is float:
        return Float(value)
    if t is str:
        return String(value)


@dataclass(unsafe_hash=True, order=False)
class Literal(ExpressionNode):
    value: Base

    @classmethod
    def create(cls, value: Base):
        tag = convert_tag(value)
        if tag is None:
            raise ValueError(f'Invalid expression node of type {type(value)} "{value}"')
        return super().create(tag)

    def __str__(self):
        return self.value.snbt()

    def __repr__(self):
        return self.value.snbt()



import re
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Callable, Iterable

from beet.core.utils import required_field
from mecha import AstCommand, AstNode, AstObjective, AstPlayerName, Reducer, rule

from .sources import ConstantScoreSource, Source

__all__ = [
    "ConstantScoreChecker",
    "ObjectiveChecker",
    "SourceJsonConverter",
]


@dataclass
class ConstantScoreChecker(Reducer):
    objective: str = required_field()
    callback: Callable = required_field()
    pattern: re.Pattern[str] = re.compile(r"^\$([-+]?\d+)\b")

    @cached_property
    def objective_node(self):
        return AstObjective(value=self.objective)

    @rule(AstCommand)
    def command(self, node: AstCommand):
        if self.objective_node not in node.arguments:
            return

        i = node.arguments.index(self.objective_node)
        name = node.arguments[i - 1]

        if not isinstance(name, AstPlayerName):
            return

        match = self.pattern.match(name.value)

        if not match:
            return

        value = int(match.group(1))
        source = ConstantScoreSource.create(value)

        self.callback(source)


@dataclass
class ObjectiveChecker(Reducer):
    whitelist: Iterable[str] = required_field()
    callback: Callable = required_field()

    @rule(AstObjective)
    def objective(self, node: AstObjective):
        value = node.value

        if value not in self.whitelist:
            return

        self.callback(value)


@dataclass
class SourceJsonConverter:
    converter: Callable[[Any, AstNode], AstNode]

    def convert(self, obj: Any):
        if isinstance(obj, Source):
            return obj.component()
        if isinstance(obj, list):
            return [self.convert(value) for value in obj]
        if isinstance(obj, dict):
            return {key: self.convert(value) for key, value in obj.items()}

        return obj

    def __call__(self, obj: Any, node: AstNode):
        obj = self.convert(obj)

        return self.converter(obj, node)



from beet import Context, Function

import bolt_expressions as expr

__all__ = [
    "beet_default",
]


def beet_default(ctx: Context):
    expr.Expression = ctx.inject(expr._Expression)  # type: ignore
    expr.Scoreboard = ctx.inject(expr._Scoreboard)  # type: ignore
    expr.Data = ctx.inject(expr._Data)  # type: ignore

    yield

    expr.Expression.generate_init()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bolt_expressions', 'bolt_expressions.contrib']

package_data = \
{'': ['*']}

install_requires = \
['beet>=0.55.0', 'bolt>=0.17.7', 'mecha>=0.59.2', 'nbtlib==1.12.1']

setup_kwargs = {
    'name': 'bolt-expressions',
    'version': '0.12.1',
    'description': 'Provides pandas-like expressions capabilities to the bolt extension of mecha',
    'long_description': '# bolt-expressions\n\n[![GitHub Actions](https://github.com/rx-modules/bolt-expressions/workflows/CI/badge.svg)](https://github.com/rx-modules/bolt-expressions/actions)\n[![PyPI](https://img.shields.io/pypi/v/bolt-expressions.svg)](https://pypi.org/project/bolt-expressions/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/bolt-expressions.svg)](https://pypi.org/project/bolt-expressions/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![Discord](https://img.shields.io/discord/900530660677156924?color=7289DA&label=discord&logo=discord&logoColor=fff)](https://discord.gg/98MdSGMm8j)\n\n> a `pandas`-esque API for creating expressions within bolt\n\n## Introduction\n\nBolt is a scripting language which mixes both python and mcfunction. This package amplifies this language by adding an API for creating fluent expressions loosely based off of the `pandas` syntax. These expressions are use for simplifying large bits of scoreboard and storage operation allowing you to swiftly create complex operations with the ease of normal programming.\n\n```py\nfrom bolt_expressions import Scoreboard\n\nmath = Scoreboard.objective("math")\n# or `math = Scoreboard("math")`\n\nmath["@s"] = math["@r"] * 10 + math["@r"] + 100\n```\n->\n```mcfunction\nscoreboard players operations $i0 bolt.expressions.temp = @r math\nscoreboard players operations $i0 bolt.expressions.temp *= #10 bolt.expressions.int\nscoreboard players operations $i1 bolt.expressions.temp = @r math\nscoreboard players add $i1 bolt.expressions.temp 100\nscoreboard players operations $i0 bolt.expressions.temp += $i1 bolt.expressions.temp\nscoreboard players operations @s math = $i0 bolt.expressions.temp\n```\n\n## Installation\n\nThe package can be installed with `pip`. Note, you must have `beet`, `mecha` and `bolt` installed to use this package.\n\n```bash\n$ pip install bolt-expressions\n```\n\n## Getting started\n\nThis package is designed to be used within any `bolt` script (either a `.mcfunction` or `bolt` file) inside a `bolt` enabled project.\n\n```yaml\nrequire:\n    - bolt\n\npipeline:\n    - mecha\n```\n\nOnce you\'ve enabled bolt, you are able to import the python package directly inside your bolt script.\n\n```py\nfrom bolt_expressions import Scoreboard, Storage\n```\n\nAny usage of the `bolt_expressions` package will require you to inject the current beet context into the API objects. Then, you can create an objective and start creating expressions.\n\n```py\nmath = Scoreboard.objective("math")\nentity_id = Scoreboard.objective("entity_id")\n\nmath["@s"] += 10\n```\n\n## Features\n\n- Robust API supporting Scoreboards, Storage, Blocks, and Entities\n- Provides an interface to manipulate large, complex mathematical expressions simplily\n- Allows you to interopt custom variables with normal commands *(soon)*\n\nCheckout some examples over at our [docs](https://rx-modules.github.io/bolt-expressions/)!\n\n## Contributing\n\nContributions are welcome. Make sure to first open an issue discussing the problem or the new feature before creating a pull request. The project uses [`poetry`](https://python-poetry.org).\n\n```bash\n$ poetry install\n```\n\nYou can run the tests with `poetry run pytest`.\n\n```bash\n$ poetry run pytest\n```\n\nThe project must type-check with [`pyright`](https://github.com/microsoft/pyright). If you\'re using VSCode the [`pylance`](https://marketplace.visualstudio.com/items?itemName=ms-python.vscode-pylance) extension should report diagnostics automatically. You can also install the type-checker locally with `npm install` and run it from the command-line.\n\n```bash\n$ npm run watch\n$ npm run check\n```\n\nThe code follows the [`black`](https://github.com/psf/black) code style. Import statements are sorted with [`isort`](https://pycqa.github.io/isort/).\n\n```bash\n$ poetry run isort bolt_expressions examples tests\n$ poetry run black bolt_expressions examples tests\n$ poetry run black --check bolt_expressions examples tests\n```\n\n---\n\nLicense - [MIT](https://github.com/rx-modules/bolt-expressions/blob/main/LICENSE)\n',
    'author': 'rx97',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rx-modules/bolt-expressions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from fractions import Fraction
from functools import wraps
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

from nbtlib import Double, Float, Numeric

from . import operations as op
from .literals import Literal
from .sources import ConstantScoreSource, DataSource, ScoreSource, TempScoreSource

# from rich import print
# from rich.pretty import pprint

__all__ = [
    "SmartGenerator",
    "Optimizer",
    "dummy",
    "noncommutative_set_collapsing",
    "commutative_set_collapsing",
    "data_set_scaling",
    "data_get_scaling",
    "literal_to_constant_replacement",
    "output_score_replacement",
    "set_to_self_removal",
    "set_and_get_cleanup",
]

if TYPE_CHECKING:
    Rule = Callable[[Iterable[op.Operation]], None]

T = TypeVar("T")


class SmartGenerator(Generator):
    """Implements `.push(val)` which allows you to 'prepend' values to a generator.
    Allows you to peek values in the future, and return them to be consumed later.

    >>> gen = SmartGenerator(i for i in range(10))
    >>> val = next(gen)
    >>> gen.push(val)
    >>> next(gen)
    0
    """

    def __init__(self, func):
        self._func: Callable[..., Generator] = func

    def __iter__(self):
        self._pre: List[T] = []
        return self

    def __next__(self) -> T:
        if self._pre:
            return self._pre.pop()

        return next(self._values)  # raises StopIteration for us

    def __call__(self, values) -> Iterable[T]:
        self._values: Iterable[T] = self._func(values)
        return self

    def push(self, val: T):
        if val != None:
            self._pre.append(val)

    def send(self, val: T):
        return self._values.send(val)

    def throw(self, val: T):
        raise self._values.throw(val)


class Optimizer:
    """Handles the operation of various optimization rules.

    The `Optimizer.rule` decorates generator functions which process Operation nodes. Each rule
    must take and return a generator of Operation nodes. This usually involves a loop which
    steps through the generator and performs optimizations across the nodes.

    Each rule is converted into a `SmartGenerator`, which adds a `.push(value)` method allowing
    you to "peek" ahead by consuming value and prepending it back to the generator. Note, calling
    `next` manually in the loop will essentially consume the node, deleting it from the final output.

    The optimizer runs the nodes through the rules like a conveyor belt and each rule acts as a worker.
    The rules can perform transformations, take nodes off and add different ones on, or even collect
    all nodes, and put on a completely different set of nodes. This metaphor should be considered when
    writing new rules.
    """

    rules: List["Rule"] = []

    @classmethod
    def rule(cls, f: Iterable["op.Operation"]):
        """Registers new rules, also converts the decorated generator into a `SmartGenerator`"""
        cls.rules.append(SmartGenerator(f))

    @classmethod
    def optimize(cls, nodes: Iterable["op.Operation"]):
        """Performs the optimization by sending all nodes through the rules."""
        for rule in cls.rules:
            nodes = rule(nodes)

        yield from nodes


@Optimizer.rule
def dummy(nodes: Iterable["op.Operation"]):
    yield from nodes


@Optimizer.rule
def noncommutative_set_collapsing(nodes: Iterable["op.Operation"]):
    """For noncommutative operations:
    ```
    scoreboard players operation $i1 temp = @s rx.uid
    scoreboard players operation $i1 temp -= $i0 temp
    scoreboard players operation @s rx.uid = $i1 temp
    ```

    ->
    ```
    scoreboard players operation @s rx.uid -= $i0 temp
    ```

    Examples to try:
    >>> abc["#value"] -= (1 + abc["@s"])    # doctest: +SKIP
    >>> abc["@s"] *= abc["@s"]              # doctest: +SKIP
    """
    for node in nodes:
        next_node = next(nodes, None)
        further_node = next(nodes, None)
        if (
            type(node) is op.Set
            and isinstance(next_node, op.ScoreOperation)
            and type(further_node) is op.Set
            and isinstance(further_node.former, ScoreSource)
            and node.latter == further_node.former
            and node.former == next_node.former
            and node.former == further_node.latter
        ):
            out = next_node.__class__(further_node.former, next_node.latter)
            yield out
        else:
            nodes.push(further_node)
            nodes.push(next_node)
            yield node


@Optimizer.rule
def commutative_set_collapsing(nodes: Iterable["op.Operation"]):
    """For commutative operations:
    ```
    scoreboard players operation $i1 temp += $i0 temp
    scoreboard players operation $i0 temp = $i1 temp
    ```
    Becomes:
    ```
    scoreboard players operation $i0 temp += $i1 temp
    ```
    """
    for node in nodes:
        next_node: Union["op.Operation", None] = next(nodes, None)
        # print("node", node)
        # print("next_node", next_node)

        if (
            type(node) in (op.Add, op.Multiply)
            and type(next_node) is op.Set
            and isinstance(next_node.former, ScoreSource)
            and node.former == next_node.latter
            and node.latter == next_node.former
        ):
            out = node.__class__(next_node.former, next_node.latter)
            # print("new", out)
            yield out
        else:
            # print("old", node)
            nodes.push(next_node)
            yield node


@Optimizer.rule
def data_set_scaling(nodes: Iterable["op.Operation"]):
    """
    Turns a multiplication/division of a temp score followed by a
    data set operation into a single set operation with a scale argument.
    ```
    scoreboard players operation $i0 temp /= $100 const
    execute store result storage demo out int 1 run scoreboard players get $i0 temp
    ```
    ->
    ```
    execute store result storage demo out float 0.01 run scoreboard players get $i0 temp
    ```

    Also works with all data operations such as append, prepend and insert:
    ```
    scoreboard players operation $i1 bolt.expr.temp *= $100 bolt.expr.const
    data modify storage demo list append value 0
    execute store result storage demo list[-1] int 1 run scoreboard players get $i1 bolt.expr.temp
    ```
    ->
    ```
    data modify storage demo list append value 0
    execute store result storage demo list[-1] int 100 run scoreboard players get $i1 bolt.expr.temp

    ```
    Examples to try:
    >>> temp.out = (obj["$value"] + 1) * 10
    >>> temp.percent = (obj["$stack"] * 100) / 64
    """
    for node in nodes:
        operation_node = None
        next_node = next(nodes, None)
        # skip data operation node (just so we can get to
        # the actual Set node)
        if isinstance(next_node, op.DataOperation):
            operation_node = next_node
            next_node = next(nodes, None)
        if (
            isinstance(node, (op.Multiply, op.Divide))
            and isinstance(next_node, op.Set)
            and isinstance(node.latter, Literal)
            and isinstance(next_node.former, DataSource)
            and node.former == next_node.latter
        ):
            scale = float(node.latter.value)

            if scale.is_integer():
                scale = int(scale)

            source = next_node.former
            number_type = source._nbt_type

            if isinstance(node, op.Divide):
                scale = 1 / scale
                number_type = number_type or source._default_floating_point_type

            new_source = source._copy(scale=scale, nbt_type=number_type)
            out = op.Set(new_source, node.former)

            if operation_node:
                yield operation_node  # yield the data operation node back in

            yield out
        else:
            nodes.push(next_node)
            nodes.push(operation_node)
            yield node


@Optimizer.rule
def data_get_scaling(nodes: Iterable["op.Operation"]):
    """
    ````
    execute store result score $i0 temp run data get storage demo value 1
    scoreboard players operation $i0 temp *= $100 const
    ```
    ->
    ```
    execute store result score $i0 temp run data get storage demo value 100
    ```
    Examples to try:
    >>> obj["#offset"] = (player.Motion[0] * 100) - obj["#x"]
    """
    for node in nodes:
        next_node = next(nodes, None)
        if (
            isinstance(node, op.Set)
            and isinstance(next_node, (op.Multiply, op.Divide))
            and isinstance(node.latter, DataSource)
            and isinstance(next_node.latter, Literal)
            and node.former == next_node.former
        ):
            scale = float(next_node.latter.value)

            if scale.is_integer():
                scale = int(scale)

            if isinstance(next_node, op.Divide):
                scale = 1 / scale

            yield op.Set(node.former, node.latter._copy(scale=scale))
        else:
            nodes.push(next_node)
            yield node


@Optimizer.rule
def multiply_divide_by_fraction(nodes: Iterable["op.Operation"]):
    for node in nodes:
        if (
            isinstance(node, (op.Multiply, op.Divide))
            and isinstance(node.latter, Literal)
            and isinstance(node.latter.value, (Float, Double))
        ):
            value = Fraction(node.latter.value).limit_denominator()

            if isinstance(node, op.Divide):
                value = 1 / value

            yield op.Multiply.create(node.former, value.numerator)
            yield op.Divide.create(node.former, value.denominator)
        else:
            yield node


@Optimizer.rule
def output_score_replacement(nodes: Iterable["op.Operation"]):
    """Replace the outermost temp score by the output score.
    If expression tree uses the output score, this rule won't be applied.
    """
    all_nodes = list(nodes)

    if all_nodes:
        # print("[bold]Applying output score replacement on tree:[/bold]")
        # pprint(all_nodes)
        last_node = all_nodes[-1]
        target_var = last_node.latter
        output_var = last_node.former
        can_replace = False
        # print(f"Last node is {last_node}")
        if (
            isinstance(output_var, ScoreSource)
            and type(last_node) is op.Set
            and len(all_nodes) > 1
        ):
            for node in all_nodes[1:]:  # Ignore the first Set operation
                # print(f"Is {node.latter} equal to {output_var}? {node.latter == output_var}")
                if node.latter == output_var:
                    break
            else:
                can_replace = True
        # print(f"Can replace output score? {can_replace}")
        def replace(source):
            if source == target_var:
                return output_var
            return source

        if can_replace:
            for node in all_nodes:
                yield node.__class__(replace(node.former), replace(node.latter))
        else:
            yield from all_nodes


@Optimizer.rule
def multiply_divide_by_one_removal(nodes: Iterable["op.Operation"]):
    for node in nodes:
        if not (
            isinstance(node, (op.Multiply, op.Divide))
            and isinstance(node.latter, Literal)
            and node.latter.value == 1
        ):
            yield node


@Optimizer.rule
def add_subtract_by_zero_removal(nodes: Iterable["op.Operation"]):
    for node in nodes:
        if not (
            isinstance(node, (op.Add, op.Subtract))
            and isinstance(node.latter, Literal)
            and node.latter.value == 0
        ):
            yield node


@Optimizer.rule
def set_to_self_removal(nodes: Iterable["op.Operation"]):
    """Removes Set operations that have the same former and latter source.
    Should run after "output_score_replacement" is applied to clean up
    all the reduntant Sets created by the previous rule.
    """
    for node in nodes:
        if type(node) is op.Set:
            if node.former != node.latter:
                yield node
        else:
            yield node


@Optimizer.rule
def set_and_get_cleanup(nodes: Iterable["op.Operation"]):
    """
    Removes unnecessary temp vars possibly originated from previous
    optimizations.
    ```
    scoreboard players operation $i0 bolt.expr.temp = $value obj
    execute store result storage demo out float 0.01 run scoreboard players get $i0 bolt.expr.temp
    ```
    ->
    ```
    execute store result storage demo out float 0.01 run scoreboard players get $value obj
    ```
    Examples to try:
    >>> temp.out = obj["$value"] / 100
    >>> temp.out = temp.value * 100 / 100
    """
    for node in nodes:
        next_node = next(nodes, None)
        if (
            isinstance(node, op.Set)
            and isinstance(next_node, op.Set)
            and node.former == next_node.latter
        ):
            out = op.Set(next_node.former, node.latter)
            yield out
        else:
            nodes.push(next_node)
            yield node


@Optimizer.rule
def literal_to_constant_replacement(
    nodes: Iterable["op.Operation"],
) -> Iterable["op.Operation"]:
    for node in nodes:
        if (
            isinstance(node, (op.Multiply, op.Divide, op.Min, op.Max, op.Modulus))
            and isinstance(node.latter, Literal)
            and isinstance(node.latter.value, Numeric)
        ):
            value = int(node.latter.value)
            constant = ConstantScoreSource.create(value)

            yield type(node).create(node.former, constant)
        else:
            yield node
